package org.xbib.jdbc.connection.pool;

import org.xbib.jdbc.connection.pool.util.BagEntry;
import org.xbib.jdbc.connection.pool.util.ClockSource;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * The house keeping task to retire and maintain minimum idle connections.
 */
final class HouseKeeper implements Runnable {

    private static final Logger logger = Logger.getLogger(HouseKeeper.class.getName());

    private final Pool pool;

    private volatile long previous;

    HouseKeeper(Pool pool) {
        this.pool = pool;
        this.previous = ClockSource.plusMillis(ClockSource.currentTime(), -pool.config.getHousekeepingPeriodMs());
    }

    @Override
    public void run() {
        try {
            logger.log(Level.FINE, () -> "housekeeper running");
            pool.connectionTimeout = pool.config.getConnectionTimeout();
            pool.validationTimeout = pool.config.getValidationTimeout();
            pool.leakTaskFactory.updateLeakDetectionThreshold(pool.config.getLeakDetectionThreshold());
            pool.catalog = (pool.config.getCatalog() != null && !pool.config.getCatalog().equals(pool.catalog)) ? pool.config.getCatalog() : pool.catalog;
            final long idleTimeout = pool.config.getIdleTimeout();
            final long now = ClockSource.currentTime();
            // allowing +128ms as per NTP spec
            if (ClockSource.plusMillis(now, 128) < ClockSource.plusMillis(previous, pool.config.getHousekeepingPeriodMs())) {
                logger.log(Level.WARNING, "retrograde clock change detected (housekeeper delta=), soft-evicting connections from pool: " +
                        pool.poolName + " " + ClockSource.elapsedDisplayString(previous, now));
                previous = now;
                pool.softEvictConnections();
                return;
            } else if (now > ClockSource.plusMillis(previous, (3 * pool.config.getHousekeepingPeriodMs()) / 2)) {
                logger.log(Level.WARNING, "thread starvation or clock leap detected: " +
                        pool.poolName + "  housekeeper delta=" + ClockSource.elapsedDisplayString(previous, now));
            }
            previous = now;
            if (idleTimeout > 0L && pool.config.getMinimumIdle() < pool.config.getMaximumPoolSize()) {
                pool.logPoolState("before cleanup");
                final List<PoolEntry> notInUse = pool.bag.values(BagEntry.STATE_NOT_IN_USE);
                int toRemove = notInUse.size() - pool.config.getMinimumIdle();
                for (PoolEntry entry : notInUse) {
                    if (toRemove > 0 && ClockSource.elapsedMillis(entry.getLastAccessed(), now) > idleTimeout && pool.bag.reserve(entry)) {
                        pool.closeConnection(entry, "(connection has passed idleTimeout)");
                        toRemove--;
                    }
                }
                pool.logPoolState("after cleanup");
            } else {
                pool.logPoolState("pool");
            }
            pool.fillPool();
        } catch (Exception e) {
            logger.log(Level.SEVERE, "unexpected exception in housekeeping task: " + e.getMessage(), e);
        }
    }
}
