package org.xbib.jdbc.mariadb.test;

import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.testcontainers.containers.MariaDBContainer;
import org.xbib.jdbc.query.Config;
import org.xbib.jdbc.query.ConfigSupplier;
import org.xbib.jdbc.query.DatabaseProvider;
import org.xbib.jdbc.query.OptionsOverride;
import org.xbib.jdbc.query.Schema;
import org.xbib.jdbc.query.Sql;
import org.xbib.jdbc.query.SqlArgs;
import org.xbib.jdbc.test.CommonTest;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static org.junit.jupiter.api.Assertions.assertEquals;

public class MariaDBTest extends CommonTest {

    static MariaDBContainer<?> mariaDBContainer;

    static {
        mariaDBContainer = new MariaDBContainer<>("mariadb")
                .withDatabaseName("testDB")
                .withUsername("testUser")
                .withPassword("testPassword");
    }

    @BeforeAll
    public static void startContainer() {
        mariaDBContainer.start();
    }

    @AfterAll
    public static void stopContainer() {
        mariaDBContainer.stop();
    }

    @Override
    protected DatabaseProvider createDatabaseProvider(OptionsOverride options) {
        Config config = ConfigSupplier.of()
                .property("database.url", mariaDBContainer.getJdbcUrl())
                .property("database.user", "testUser")
                .property("database.password", "testPassword")
                .get();
        return DatabaseProvider.builder(config)
                .withSqlParameterLogging()
                .withSqlInExceptionMessages()
                .withOptions(options)
                .build();
    }

    @Disabled("MariaDB prohibits NaN and Infinity")
    @Test
    public void argFloatNaN() {
        super.argFloatNaN();
    }

    @Disabled("MariaDB prohibits NaN and Infinity")
    @Test
    public void argFloatInfinity() {
        super.argFloatInfinity();
    }

    @Disabled("MariaDB prohibits NaN and Infinity")
    @Test
    public void argDoubleNaN() {
        super.argDoubleNaN();
    }

    @Disabled("MariaDB prohibits NaN and Infinity")
    @Test
    public void argDoubleInfinity() {
        super.argDoubleInfinity();
    }

    @Disabled("MariaDB prohibits NaN and Infinity")
    @Test
    public void argFloatNegativeZero() {
        super.argFloatNegativeZero();
    }

    @Disabled("MariaDB prohibits NaN and Infinity")
    @Test
    public void argDoubleNegativeZero() {
        super.argDoubleNegativeZero();
    }

    @Disabled("MariaDB temporarily disabled")
    @Test
    @Override
    public void intervals() {
        super.intervals();
    }

    @Disabled("MariaDB temporarily disabled")
    @Test
    @Override
    public void metadataColumnTypes() {
        super.metadataColumnTypes();
    }

    /**
     * MariaDB seems to have different behavior in that is does not convert
     * column names.
     * I haven't figured out how to smooth over this difference, since all databases
     * seem to respect the provided case when it is inside quotes, but don't provide
     * a way to tell whether a particular parameter was quoted.
     */
    @Override
    @Test
    public void metadataColumnNames() {
        db.dropTableQuietly("dbtest");
        new Schema().addTable("dbtest").addColumn("pk").primaryKey().schema().execute(db);
        db.toSelect("select Pk, Pk as Foo, Pk as \"Foo\" from dbtest")
                .query(rs -> {
                    Assertions.assertArrayEquals(new String[]{"Pk", "Foo", "Foo"}, rs.getColumnLabels());
                    return null;
                });
    }

    /**
     * This one is adjusted in that the float values are passed as double, because
     * the database stores them both as double and there doesn't appear to be a way
     * to tell that one was actually declared as a float.
     *
     * Also, CLOBs are stored as strings ("mediumtext" is just an alias).
     */
    @Test
    public void saveResultAsTable() {
        new Schema().addTable("dbtest")
                .addColumn("nbr_integer").asInteger().primaryKey().table()
                .addColumn("nbr_long").asLong().table()
                .addColumn("nbr_float").asFloat().table()
                .addColumn("nbr_double").asDouble().table()
                .addColumn("nbr_big_decimal").asBigDecimal(19, 9).table()
                .addColumn("str_varchar").asString(80).table()
                .addColumn("str_fixed").asStringFixed(1).table()
                .addColumn("str_lob").asClob().table()
                .addColumn("bin_blob").asBlob().table()
                .addColumn("boolean_flag").asBoolean().table()
                .addColumn("date_millis").asLocalDateTime().table()
                .addColumn("local_date").asLocalDate().schema().execute(db);
        db.toInsert("insert into dbtest (nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal, str_varchar,"
                        + " str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date) values (?,?,?,?,?,?,?,?,?,?,?,?)")
                .argInteger(Integer.MAX_VALUE)
                .argLong(Long.MAX_VALUE)
                .argDouble((double) Float.MAX_VALUE)
                .argDouble(Double.MAX_VALUE)
                .argBigDecimal(new BigDecimal("123.456"))
                .argString("hello")
                .argString("Z")
                .argClobString("hello again")
                .argBlobBytes(new byte[]{'1', '2'})
                .argBoolean(true)
                .argLocalDateTime(now)
                .argLocalDate(localDateNow)
                .insert(1);

        db.toInsert("insert into dbtest (nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal, str_varchar,"
                        + " str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date) values (?,?,?,?,?,?,?,?,?,?,?,?)")
                .argInteger(Integer.MIN_VALUE)
                .argLong(Long.MIN_VALUE)
                .argDouble((double) Float.MIN_VALUE)
                .argDouble(Double.MIN_VALUE)
                .argBigDecimal(new BigDecimal("-123.456"))
                .argString("goodbye")
                .argString("A")
                .argClobString("bye again")
                .argBlobBytes(new byte[]{'3', '4'})
                .argBoolean(false)
                .argLocalDateTime(now)
                .argLocalDate(localDateNow)
                .insert(1);

        String expectedSchema = new Schema().addTable("dbtest2")
                .addColumn("nbr_integer").asInteger().table()
                .addColumn("nbr_long").asLong().table()
                .addColumn("nbr_float").asFloat().table()
                .addColumn("nbr_double").asDouble().table()
                .addColumn("nbr_big_decimal").asBigDecimal(19, 9).table()
                .addColumn("str_varchar").asString(80).table()
                .addColumn("str_fixed").asStringFixed(1).table()
                .addColumn("str_lob").asClob().table()
                .addColumn("bin_blob").asBlob().table()
                .addColumn("boolean_flag").asBoolean().table()
                .addColumn("date_millis").asLocalDateTime().table()
                .addColumn("local_date").asLocalDate()
                .schema()
                .print(db.flavor());

        List<SqlArgs> args = db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                        + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest")
                .query(rs -> {
                    List<SqlArgs> result = new ArrayList<>();
                    while (rs.next()) {
                        if (result.size() == 0) {
                            db.dropTableQuietly("dbtest2");
                            Schema schema = new Schema().addTableFromRow("dbtest2", rs).schema();
                            assertEquals(expectedSchema, schema.print(db.flavor()));
                            schema.execute(db);
                        }
                        result.add(SqlArgs.readRow(rs));
                    }
                    return result;
                });

        db.toInsert(Sql.insert("dbtest2", args)).insertBatch();

        assertEquals(2, db.toSelect("select count(*) from dbtest2").queryIntegerOrZero());
        assertEquals(db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                                + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest order by 1")
                        .queryMany(SqlArgs::readRow),
                db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                                + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest2 order by 1")
                        .queryMany(SqlArgs::readRow));
        assertEquals(Arrays.asList(
                        new SqlArgs()
                                .argInteger("nbr_integer", Integer.MIN_VALUE)
                                .argLong("nbr_long", Long.MIN_VALUE)
                                .argDouble("nbr_float", (double) Float.MIN_VALUE)
                                .argDouble("nbr_double", Double.MIN_VALUE)
                                .argBigDecimal("nbr_big_decimal", new BigDecimal("-123.456"))
                                .argString("str_varchar", "goodbye")
                                .argString("str_fixed", "A")
                                .argString("str_lob", "bye again")
                                .argBlobBytes("bin_blob", new byte[]{'3', '4'})
                                .argString("boolean_flag", "N")//.argBoolean("boolean_flag", false)
                                .argLocalDateTime("date_millis", now)
                                .argLocalDate("local_date", localDateNow),
                        new SqlArgs()
                                .argInteger("nbr_integer", Integer.MAX_VALUE)
                                .argLong("nbr_long", Long.MAX_VALUE)
                                .argDouble("nbr_float", (double) Float.MAX_VALUE)
                                .argDouble("nbr_double", Double.MAX_VALUE)
                                .argBigDecimal("nbr_big_decimal", new BigDecimal("123.456"))
                                .argString("str_varchar", "hello")
                                .argString("str_fixed", "Z")
                                .argString("str_lob", "hello again")
                                .argBlobBytes("bin_blob", new byte[]{'1', '2'})
                                .argString("boolean_flag", "Y")//.argBoolean("boolean_flag", true)
                                .argLocalDateTime("date_millis", now)
                                .argLocalDate("local_date", localDateNow)),
                db.toSelect("select nbr_integer, nbr_long, nbr_float, nbr_double, nbr_big_decimal,"
                                + " str_varchar, str_fixed, str_lob, bin_blob, boolean_flag, date_millis, local_date from dbtest2 order by 1")
                        .queryMany(SqlArgs::readRow));
    }
}
