package org.xbib.jdbc.query;

import org.xbib.jdbc.connection.pool.PoolConfig;
import org.xbib.jdbc.connection.pool.PoolDataSource;
import org.xbib.jdbc.query.util.Metric;

import javax.sql.DataSource;
import java.io.Closeable;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * This is a lazy provider for Database instances. It helps avoid allocating connection
 * or transaction resources until (or if) we actually need a Database. As a consequence
 * of this laziness, the underlying resources require explicit cleanup by calling either
 * commitAndClose() or rollbackAndClose().
 */
public final class DatabaseProvider implements Supplier<Database>, Closeable {
    
    private static final Logger logger = Logger.getLogger(DatabaseProvider.class.getName());
    
    private static final AtomicInteger poolNameCounter = new AtomicInteger(1);

    private final DatabaseProviderBuilderImpl builder;

    private Connection connection;
    
    private Database database;

    private DatabaseProvider(DatabaseProviderBuilderImpl builder) {
        this.builder = builder;
    }

    /**
     * Use an externally configured DataSource and a Flavor.
     */
    public static DatabaseProviderBuilder builder(DataSource ds,
                                                  String flavor) {
        return new DatabaseProviderBuilderImpl(ds, () -> {
            try {
                if (ds == null) {
                    throw new NullPointerException();
                }
                return ds.getConnection();
            } catch (Exception e) {
                throw new DatabaseException("unable to obtain a connection from the DataSource", e);
            }
        }, new OptionsDefault(Flavor.valueOf(flavor)));
    }

    /**
     * Configure the database from the following properties read from the provided configuration:
     * <pre>
     *   database.url=...       Database connect string (required)
     *   database.user=...      Authenticate as this user (optional if provided in url)
     *   database.password=...  User password (optional if user and password provided in
     *                          url; prompted on standard input if user is provided and
     *                          password is not)
     *   database.pool.size=... How many connections in the connection pool (default 10).
     *   database.driver.class  The driver to initialize with Class.forName(). This will
     *                          be guessed from the database.url if not provided.
     *   database.flavor        A {@link Flavor}. If this is not provided the flavor will be guessed based on the
     *                          value for database.url, if possible.
     * </pre>
     *
     * <p>The database flavor will be guessed based on the URL.</p>
     *
     * <p>A database pool will be created using jdbc-connection-pool.</p>
     */
    public static DatabaseProviderBuilder builder(Config config) {
        String flavorString = config.getString("database.flavor");
        if (flavorString == null) {
            String url = config.getString("database.url");
            if (url == null) {
                throw new DatabaseException("You must provide database.url");
            }
            flavorString = Flavor.fromJdbcUrl(url).getName();
        }
        return builder(createDataSource(config), flavorString);
    }

    /**
     * Builder method to create and initialize an instance of this class using
     * the JDBC standard DriverManager method. The url parameter will be inspected
     * to determine the Flavor for this database. Short version for
     * anonymous access without user name or password.
     */
    public static DatabaseProviderBuilder builder(ClassLoader classLoader,
                                                  String url) {
        return builder(classLoader, url, Flavor.fromJdbcUrl(url).driverClass(), null, null, null);
    }

    /**
     * Builder method to create and initialize an instance of this class using
     * the JDBC standard DriverManager method.
     * @param classLoader the classloader
     * @param url the database URL
     * @param driverClass the driver class
     * @param properties the driver properties or null
     * @param user user name or null
     * @param password user password or null
     * @return a database provider builder
     */
    public static DatabaseProviderBuilder builder(ClassLoader classLoader,
                                                  String url,
                                                  String driverClass,
                                                  Properties properties,
                                                  String user,
                                                  String password) {
        try {
            DriverManager.getDriver(url);
        } catch (SQLException e) {
            if (driverClass != null) {
                try {
                    Class.forName(driverClass, true, classLoader);
                } catch (ClassNotFoundException e1) {
                    throw new DatabaseException("couldn't locate JDBC driver", e1);
                }
            }
        }
        return new DatabaseProviderBuilderImpl(null, () -> {
            try {
                if (properties != null) {
                    return DriverManager.getConnection(url, properties);
                } else if (user != null) {
                    return DriverManager.getConnection(url, user, password);
                }
                return DriverManager.getConnection(url);
            } catch (Exception e) {
                throw new DatabaseException("unable to obtain a connection from DriverManager", e);
            }
        }, new OptionsDefault(Flavor.fromJdbcUrl(url)));
    }

    private static DataSource createDataSource(Config config) {
        String url = config.getString("database.url");
        if (url == null) {
            throw new DatabaseException("You must provide database.url");
        }
        PoolConfig poolConfig = new PoolConfig();
        poolConfig.setPoolName(config.getString("database.pool.name", "database-pool-" + poolNameCounter.getAndAdd(1)));
        poolConfig.setUrl(config.getString("database.url"));
        // provide driver class only if in config
        String driverClassName = config.getString("database.driver.class");
        if (driverClassName != null) {
            poolConfig.setDriverClassName(config.getString("database.driver.class"));
        }
        poolConfig.setUsername(config.getString("database.user"));
        poolConfig.setPassword(config.getString("database.password"));
        poolConfig.setMaximumPoolSize(config.getInteger("database.pool.size", 8));
        poolConfig.setAutoCommit(false);
        try {
            return new PoolDataSource(poolConfig);
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            return null;
        }
    }

    public void transact(final DbCode code) {
        boolean complete = false;
        try {
            code.run(this);
            complete = true;
        } catch (ThreadDeath | DatabaseException t) {
            throw t;
        } catch (Throwable t) {
            throw new DatabaseException("Error during transaction", t);
        } finally {
            if (!complete) {
                rollback();
            } else {
                commit();
            }
        }
    }

    public void transactAndClose(final DbCode code) {
        boolean complete = false;
        try {
            code.run(this);
            complete = true;
        } catch (ThreadDeath | DatabaseException t) {
            throw t;
        } catch (Throwable t) {
            throw new DatabaseException("Error during transaction", t);
        } finally {
            if (!complete) {
                rollbackAndClose();
            } else {
                commitAndClose();
            }
        }
    }

    public <T> T transactReturning(final DbCodeTyped<T> code) {
        T result;
        boolean complete = false;
        try {
            result = code.run(this);
            complete = true;
        } catch (ThreadDeath | DatabaseException t) {
            throw t;
        } catch (Throwable t) {
            throw new DatabaseException("Error during transaction", t);
        } finally {
            if (!complete) {
                rollback();
            } else {
                commit();
            }
        }
        return result;
    }

    public <T> T transactReturningAndClose(final DbCodeTyped<T> code) {
        T result;
        boolean complete = false;
        try {
            result = code.run(this);
            complete = true;
        } catch (ThreadDeath | DatabaseException t) {
            throw t;
        } catch (Throwable t) {
            throw new DatabaseException("Error during transaction", t);
        } finally {
            if (!complete) {
                rollbackAndClose();
            } else {
                commitAndClose();
            }
        }
        return result;
    }

    public void transact(final DbCodeTx code) {
        Transaction tx = new TransactionImpl();
        tx.setRollbackOnError(true);
        tx.setRollbackOnly(false);
        boolean complete = false;
        try {
            code.run(this, tx);
            complete = true;
        } catch (ThreadDeath | DatabaseException t) {
            throw t;
        } catch (Throwable t) {
            throw new DatabaseException("Error during transaction", t);
        } finally {
            if ((!complete && tx.isRollbackOnError()) || tx.isRollbackOnly()) {
                rollback();
            } else {
                commit();
            }
        }
    }

    public void transactAndClose(final DbCodeTx code) {
        Transaction tx = new TransactionImpl();
        tx.setRollbackOnError(true);
        tx.setRollbackOnly(false);
        boolean complete = false;
        try {
            code.run(this, tx);
            complete = true;
        } catch (ThreadDeath | DatabaseException t) {
            throw t;
        } catch (Throwable t) {
            throw new DatabaseException("Error during transaction", t);
        } finally {
            if ((!complete && tx.isRollbackOnError()) || tx.isRollbackOnly()) {
                rollbackAndClose();
            } else {
                commitAndClose();
            }
        }
    }

    public Database get() {
        if (database != null) {
            return database;
        }
        if (builder.isClosed()) {
            throw new DatabaseException("Called get() on a DatabaseProvider after close()");
        }
        Metric metric = new Metric(logger.isLoggable(Level.FINE));
        try {
            connection = builder.connectionProvider.get();
            metric.checkpoint("getConn");
            try {
                // JDBC specifies that autoCommit is the default for all new connections.
                // Don't try to be clever about clearing it conditionally.
                if (!builder.options.flavor().isAutoCommitOnly()) {
                    connection.setAutoCommit(false);
                    metric.checkpoint("setAutoCommit");
                }
            } catch (SQLException e) {
                throw new DatabaseException("Unable to set autoCommit for the connection", e);
            }
            database = new DatabaseImpl(connection, builder.options);
            metric.checkpoint("dbInit");
        } catch (RuntimeException e) {
            metric.checkpoint("fail");
            throw e;
        } finally {
            metric.done();
            if (logger.isLoggable(Level.FINE)) {
                StringBuilder sb = new StringBuilder("Get ").append(builder.options.flavor()).append(" database: ");
                metric.printMessage(sb);
                logger.fine(sb.toString());
            }
        }
        return database;
    }

    public boolean check(int seconds) {
        Metric metric = new Metric(logger.isLoggable(Level.FINE));
        boolean probe = false;
        try {
            if (database != null) {
                probe = database.probe(seconds);
            } else {
                get();
                probe = database.probe(seconds);
            }
        } catch (SQLException | RuntimeException e) {
            metric.checkpoint("fail", e.getMessage());
        } finally {
            metric.done();
            if (logger.isLoggable(Level.FINE)) {
                StringBuilder sb = new StringBuilder("Probe ").append(builder.options.flavor()).append(" database: ");
                metric.printMessage(sb);
                logger.log(Level.FINE, sb.toString());
            }
        }
        return probe;
    }

    public void commit() {
        if (connection != null) {
            try {
                if (!builder.options.flavor().isAutoCommitOnly()) {
                    connection.commit();
                }
            } catch (Exception e) {
                throw new DatabaseException("Unable to commit the transaction", e);
            }
        }
    }

    public void commitAndClose() {
        try {
            if (connection != null) {
                if (!builder.options.flavor().isAutoCommitOnly()) {
                    connection.commit();
                }
            }
            close();
        } catch (Exception e) {
            throw new DatabaseException("Unable to commit the transaction", e);
        }
    }

    public void rollback() {
        if (connection != null) {
            try {
                if (!builder.options.flavor().isAutoCommitOnly()) {
                    connection.rollback();
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE, "Unable to rollback the transaction", e);
            }
        }
    }

    public void rollbackAndClose() {
        try {
            if (connection != null) {
                if (!builder.options.flavor().isAutoCommitOnly()) {
                    connection.rollback();
                }
            }
            close();
        } catch (Exception e) {
            logger.log(Level.SEVERE, "Unable to rollback the transaction", e);
        }
    }

    @Override
    public void close() throws IOException {
        if (connection != null) {
            try {
                connection.close();
            } catch (Exception e) {
                logger.log(Level.SEVERE, "Unable to close the database connection", e);
            }
        }
        connection = null;
        database = null;
        builder.close();
    }

    private static class DatabaseProviderBuilderImpl implements DatabaseProviderBuilder, Closeable {

        private DataSource dataSource;

        private final Supplier<Connection> connectionProvider;

        private final Options options;

        private final AtomicBoolean closed;

        private DatabaseProviderBuilderImpl(DataSource dataSource,
                                            Supplier<Connection> connectionProvider,
                                            Options options) {
            this.dataSource = dataSource;
            this.connectionProvider = connectionProvider;
            this.options = options;
            this.closed = new AtomicBoolean(false);
        }

        @Override
        public DatabaseProviderBuilder withOptions(OptionsOverride options) {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, options.withParent(this.options));
        }

        @Override
        public DatabaseProviderBuilder withSqlParameterLogging() {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, new OptionsOverride() {
                @Override
                public boolean isLogParameters() {
                    return true;
                }
            }.withParent(this.options));
        }

        @Override
        public DatabaseProviderBuilder withSqlInExceptionMessages() {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, new OptionsOverride() {
                @Override
                public boolean isDetailedExceptions() {
                    return true;
                }
            }.withParent(this.options));
        }

        @Override
        public DatabaseProviderBuilder withDatePerAppOnly() {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, new OptionsOverride() {
                @Override
                public boolean useLocalDateTimeOnly() {
                    return true;
                }
            }.withParent(this.options));
        }

        @Override
        public DatabaseProviderBuilder withTransactionControl() {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, new OptionsOverride() {
                @Override
                public boolean allowTransactionControl() {
                    return true;
                }
            }.withParent(this.options));
        }

        @Override
        public DatabaseProviderBuilder withTransactionControlSilentlyIgnored() {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, new OptionsOverride() {
                @Override
                public boolean ignoreTransactionControl() {
                    return true;
                }
            }.withParent(this.options));
        }

        @Override
        public DatabaseProviderBuilder withConnectionAccess() {
            return new DatabaseProviderBuilderImpl(dataSource, connectionProvider, new OptionsOverride() {
                @Override
                public boolean allowConnectionAccess() {
                    return true;
                }
            }.withParent(this.options));
        }

        @Override
        public DatabaseProvider build() {
            return new DatabaseProvider(this);
        }

        @Override
        public void transact(DbCode tx) {
            this.build().transact(tx);
        }

        @Override
        public <T> T transactReturning(DbCodeTyped<T> tx) {
            return this.build().transactReturning(tx);
        }

        @Override
        public void transact(DbCodeTx tx) {
            this.build().transact(tx);
        }

        @Override
        public void close() throws IOException {
            if (closed.compareAndSet(false, true)) {
                if (dataSource != null) {
                    if (dataSource instanceof Closeable) {
                        ((Closeable) dataSource).close();
                    }
                    dataSource = null;
                }
            }
        }

        public boolean isClosed() {
            return closed.get();
        }
    }
}
