package org.xbib.jdbc.query;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.sql.ParameterMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.sql.Types;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

/**
 * Deal with mapping parameters into prepared statements.
 */
public class StatementAdapter {

    private static final Logger logger = Logger.getLogger(StatementAdapter.class.getName());

    private final Options options;

    public StatementAdapter(Options options) {
        this.options = options;
    }

    public void addParameters(PreparedStatement ps, Object[] parameters) throws SQLException {
        for (int i = 0; i < parameters.length; i++) {
            Object parameter = parameters[i];
            // Unwrap secret args here so we can use them
            if (parameter instanceof SecretArg) {
                parameter = ((SecretArg) parameter).getArg();
            }
            if (parameter == null) {
                ParameterMetaData metaData;
                int parameterType;
                try {
                    metaData = ps.getParameterMetaData();
                    parameterType = metaData.getParameterType(i + 1);
                } catch (SQLException e) {
                    throw new DatabaseException("Parameter " + (i + 1)
                            + " was null and the JDBC driver could not report the type of this column."
                            + " Please update the JDBC driver to support PreparedStatement.getParameterMetaData()"
                            + " or use SqlNull in place of null values to this query.", e);
                }
                ps.setNull(i + 1, parameterType);
            } else if (parameter instanceof SqlNull) {
                SqlNull sqlNull = (SqlNull) parameter;
                if (options.useBytesForBlob() && sqlNull.getType() == Types.BLOB) {
                    // The setNull() seems more correct, but PostgreSQL chokes on it
                    ps.setBytes(i + 1, null);
                } else {
                    ps.setNull(i + 1, sqlNull.getType());
                }
            } else if (parameter instanceof java.sql.Date) {
                ps.setDate(i + 1, (java.sql.Date) parameter);
            } else if (parameter instanceof Date) {
                Date date = (Date) parameter;
                LocalDateTime localDateTime = LocalDateTime.ofInstant(date.toInstant(), ZoneId.systemDefault());
                ps.setTimestamp(i + 1, Timestamp.valueOf(localDateTime));
                // this will correct the millis and nanos according to the JDBC spec
                // if a correct Timestamp is passed in, this will detect that and leave it alone
                //ps.setTimestamp(i + 1, toSqlTimestamp((Date) parameter), options.calendarForTimestamps());
            } else if (parameter instanceof Reader) {
                if (options.useStringForClob()) {
                    try (BufferedReader reader = new BufferedReader((Reader) parameter)) {
                        ps.setString(i + 1, reader.lines().collect(Collectors.joining()));
                    } catch (IOException e) {
                        throw new SQLException(e);
                    }
                } else {
                    ps.setCharacterStream(i + 1, (Reader) parameter);
                }
            } else if (parameter instanceof InputStream) {
                if (options.useBytesForBlob()) {
                    try {
                        ps.setBytes(i + 1, ((InputStream) parameter).readAllBytes());
                    } catch (IOException e) {
                        throw new SQLException(e);
                    }
                } else {
                    ps.setBinaryStream(i + 1, (InputStream) parameter);
                }
            } else if (parameter instanceof Float) {
                options.flavor().setFloat(ps, i + 1, (Float) parameter);
            } else if (parameter instanceof Double) {
                options.flavor().setDouble(ps, i + 1, (Double) parameter);
            } else {
                ps.setObject(i + 1, parameter);
            }
        }
    }

    public Object nullLocalDateTime(LocalDateTime arg) {
        return arg == null ? new SqlNull(Types.TIMESTAMP) : Timestamp.valueOf(arg);
    }

    public Object nullLocalDateTime(LocalDateTime arg, ZoneId zoneId) {
        return arg == null ? new SqlNull(Types.TIMESTAMP) : Timestamp.valueOf(arg.atZone(zoneId).toLocalDateTime());
    }

    public Object nullLocalDate(LocalDate arg) {
        return arg == null ? new SqlNull(Types.DATE) : java.sql.Date.valueOf(arg);
    }

    public Object nullNumeric(Number arg) {
        if (arg == null) {
            return new SqlNull(Types.NUMERIC);
        }
        return arg;
    }

    public Object nullString(String arg) {
        if (arg == null) {
            return new SqlNull(Types.VARCHAR);
        }
        return arg;
    }

    public Object nullClobReader(Reader arg) {
        if (arg == null) {
            return new SqlNull(Types.VARCHAR);
        }
        return arg;
    }

    public Object nullBytes(byte[] arg) {
        if (arg == null) {
            return new SqlNull(Types.BLOB);
        }
        return arg;
    }

    public Object nullInputStream(InputStream arg) {
        if (arg == null) {
            return new SqlNull(Types.BLOB);
        }
        return arg;
    }

    public void closeQuietly(ResultSet rs, Logger log) {
        if (rs != null) {
            try {
                rs.close();
            } catch (Exception e) {
                if (log != null) {
                    log.log(Level.SEVERE, "Caught exception closing the ResultSet", e);
                }
            }
        }
    }

    public void closeQuietly( Statement s,  Logger log) {
        if (s != null) {
            try {
                s.close();
            } catch (Exception e) {
                if (log != null) {
                    log.log(Level.SEVERE, "Caught exception closing the Statement", e);
                }
            }
        }
    }
}
